@@ -27,14 +27,13 @@
 
 #include "awesome.h"
 #include "event.h"
-#include "tag.h"
-#include "window.h"
 #include "property.h"
+#include "objects/tag.h"
+#include "objects/drawin.h"
+#include "xwindow.h"
 #include "ewmh.h"
-#include "client.h"
-#include "widget.h"
-#include "titlebar.h"
-#include "key.h"
+#include "objects/client.h"
+#include "keyresolv.h"
 #include "keygrabber.h"
 #include "mousegrabber.h"
 #include "luaa.h"
@@ -128,16 +127,43 @@ event_handle_mousegrabber(int x, int y, uint16_t mask)
     return false;
 }
 
+/** Emit a button signal.
+ * The top of the lua stack has to be the object on which to emit the event.
+ * \param ev The event to handle.
+ */
+static void
+event_emit_button(xcb_button_press_event_t *ev)
+{
+    const char *name;
+    switch(XCB_EVENT_RESPONSE_TYPE(ev))
+    {
+    case XCB_BUTTON_PRESS:
+        name = "button::press";
+        break;
+    case XCB_BUTTON_RELEASE:
+        name = "button::release";
+        break;
+    default:
+        fatal("Invalid event type");
+    }
+
+    /* Push the event's info */
+    lua_pushnumber(globalconf.L, ev->event_x);
+    lua_pushnumber(globalconf.L, ev->event_y);
+    lua_pushnumber(globalconf.L, ev->detail);
+    luaA_pushmodifiers(globalconf.L, ev->state);
+    /* And emit the signal */
+    luaA_object_emit_signal(globalconf.L, -5, name, 4);
+}
+
 /** The button press event handler.
  * \param ev The event.
  */
 static void
 event_handle_button(xcb_button_press_event_t *ev)
 {
-    int screen;
-    const int nb_screen = xcb_setup_roots_length(xcb_get_setup(globalconf.connection));
     client_t *c;
-    wibox_t *wibox;
+    drawin_t *drawin;
 
     globalconf.timestamp = ev->time;
 
@@ -150,57 +176,56 @@ event_handle_button(xcb_button_press_event_t *ev)
      * drop them */
     ev->state &= 0x00ff;
 
-    if((wibox = wibox_getbywin(ev->event))
-       || (wibox = wibox_getbywin(ev->child)))
+    if((drawin = drawin_getbywin(ev->event))
+       || (drawin = drawin_getbywin(ev->child)))
     {
-        /* If the wibox is child, then x,y are
+        /* If the drawin is child, then x,y are
          * relative to root window */
-        if(wibox->window == ev->child)
+        if(drawin->window == ev->child)
         {
-            ev->event_x -= wibox->geometry.x;
-            ev->event_y -= wibox->geometry.y;
+            ev->event_x -= drawin->geometry.x + drawin->border_width;
+            ev->event_y -= drawin->geometry.y + drawin->border_width;
         }
 
-        /* Push the wibox */
-        luaA_object_push(globalconf.L, wibox);
-        /* Duplicate the wibox */
-        lua_pushvalue(globalconf.L, -1);
-        /* Handle the button event on it */
-        event_button_callback(ev, &wibox->buttons, -1, 1, NULL);
-
-        /* then try to match a widget binding */
-        widget_t *w = widget_getbycoords(wibox->orientation, &wibox->widgets,
-                                         wibox->geometry.width,
-                                         wibox->geometry.height,
-                                         &ev->event_x, &ev->event_y);
-        if(w)
-        {
-            /* Push widget (the wibox is already on stack) */
-            luaA_object_push_item(globalconf.L, -1, w);
-            /* Move widget before wibox */
-            lua_insert(globalconf.L, -2);
-            event_button_callback(ev, &w->buttons, -2, 2, NULL);
-        }
-        else
-            /* Remove the wibox, we did not use it */
-            lua_pop(globalconf.L, 1);
-
+        /* Push the drawable */
+        luaA_object_push(globalconf.L, drawin);
+        luaA_object_push_item(globalconf.L, -1, drawin->drawable);
+        /* and handle the button raw button event */
+        event_emit_button(ev);
+        lua_pop(globalconf.L, 1);
+        /* check if any button object matches */
+        event_button_callback(ev, &drawin->buttons, -1, 1, NULL);
     }
-    else if((c = client_getbywin(ev->event)))
+    else if((c = client_getbyframewin(ev->event)))
     {
         luaA_object_push(globalconf.L, c);
+        /* And handle the button raw button event */
+        event_emit_button(ev);
+        /* then check if a titlebar was "hit" */
+        int x = ev->event_x, y = ev->event_y;
+        drawable_t *d = client_get_drawable_offset(c, &x, &y);
+        if (d)
+        {
+            /* Copy the event so that we can fake x/y */
+            xcb_button_press_event_t event = *ev;
+            event.event_x = x;
+            event.event_y = y;
+            luaA_object_push_item(globalconf.L, -1, d);
+            event_emit_button(&event);
+            lua_pop(globalconf.L, 1);
+        }
+        /* then check if any button objects match */
         event_button_callback(ev, &c->buttons, -1, 1, NULL);
         xcb_allow_events(globalconf.connection,
                          XCB_ALLOW_REPLAY_POINTER,
                          XCB_CURRENT_TIME);
     }
     else if(ev->child == XCB_NONE)
-        for(screen = 0; screen < nb_screen; screen++)
-            if(xutil_screen_get(globalconf.connection, screen)->root == ev->event)
-            {
-                event_button_callback(ev, &globalconf.buttons, 0, 0, NULL);
-                return;
-            }
+        if(globalconf.screen->root == ev->event)
+        {
+            event_button_callback(ev, &globalconf.buttons, 0, 0, NULL);
+            return;
+        }
 }
 
 static void
@@ -256,11 +281,10 @@ static void
 event_handle_configurerequest(xcb_configure_request_event_t *ev)
 {
     client_t *c;
-    area_t geometry;
 
     if((c = client_getbywin(ev->window)))
     {
-        geometry = titlebar_geometry_remove(c->titlebar, c->border_width, c->geometry);
+        area_t geometry = c->geometry;
 
         if(ev->value_mask & XCB_CONFIG_WINDOW_X)
             geometry.x = ev->x;
@@ -274,24 +298,13 @@ event_handle_configurerequest(xcb_configure_request_event_t *ev)
         if(ev->value_mask & XCB_CONFIG_WINDOW_BORDER_WIDTH)
         {
             luaA_object_push(globalconf.L, c);
-            client_set_border_width(globalconf.L, -1, ev->border_width);
+            window_set_border_width(globalconf.L, -1, ev->border_width);
             lua_pop(globalconf.L, 1);
         }
 
-        /* Clients are not allowed to directly mess with stacking parameters. */
-        ev->value_mask &= ~(XCB_CONFIG_WINDOW_SIBLING |
-                                            XCB_CONFIG_WINDOW_STACK_MODE);
-
-        /** Configure request are sent with size relative to real (internal)
-         * window size, i.e. without titlebars and borders. */
-        geometry = titlebar_geometry_add(c->titlebar, c->border_width, geometry);
-
         if(!client_resize(c, geometry, false))
-        {
-            /* Resize wasn't officially needed, but we don't want to break expectations. */
-            geometry = titlebar_geometry_remove(c->titlebar, c->border_width, c->geometry);
-            window_configure(c->window, geometry, c->border_width);
-        }
+            /* ICCCM 4.1.5 / 4.2.3, if nothing was changed, send an event saying so */
+            client_send_configure(c);
     }
     else
         event_handle_configurerequest_configure_window(ev);
@@ -303,16 +316,13 @@ event_handle_configurerequest(xcb_configure_request_event_t *ev)
 static void
 event_handle_configurenotify(xcb_configure_notify_event_t *ev)
 {
-    int screen_nbr;
-    const xcb_screen_t *screen;
+    const xcb_screen_t *screen = globalconf.screen;
 
-    for(screen_nbr = 0; screen_nbr < xcb_setup_roots_length(xcb_get_setup(globalconf.connection)); screen_nbr++)
-        if((screen = xutil_screen_get(globalconf.connection, screen_nbr)) != NULL
-           && ev->window == screen->root
-           && (ev->width != screen->width_in_pixels
-               || ev->height != screen->height_in_pixels))
-            /* it's not that we panic, but restart */
-            awesome_restart();
+    if(ev->window == screen->root
+       && (ev->width != screen->width_in_pixels
+           || ev->height != screen->height_in_pixels))
+        /* it's not that we panic, but restart */
+        awesome_restart();
 }
 
 /** The destroy notify event handler.
@@ -324,97 +334,59 @@ event_handle_destroynotify(xcb_destroy_notify_event_t *ev)
     client_t *c;
 
     if((c = client_getbywin(ev->window)))
-        client_unmanage(c);
+        client_unmanage(c, false);
     else
         for(int i = 0; i < globalconf.embedded.len; i++)
             if(globalconf.embedded.tab[i].win == ev->window)
             {
                 xembed_window_array_take(&globalconf.embedded, i);
-                widget_invalidate_bytype(widget_systray);
+                luaA_systray_invalidate();
             }
 }
 
-/** Handle a motion notify over widgets.
- * \param wibox The wibox.
- * \param mouse_over The pointer to the registered mouse over widget.
- * \param widget The new widget the mouse is over.
- */
-static void
-event_handle_widget_motionnotify(wibox_t *wibox,
-                                 widget_t *widget)
-{
-    if(widget != wibox->mouse_over)
-    {
-        if(wibox->mouse_over)
-        {
-            /* Emit mouse leave signal on old widget:
-             * - Push wibox.*/
-            luaA_object_push(globalconf.L, wibox);
-            /* - Push the widget the mouse was on */
-            luaA_object_push_item(globalconf.L, -1, wibox->mouse_over);
-            /* - Invert wibox/widget */
-            lua_insert(globalconf.L, -2);
-            /* - Emit the signal mouse::leave with the wibox as argument */
-            luaA_object_emit_signal(globalconf.L, -2, "mouse::leave", 1);
-            /* - Remove the widget */
-            lua_pop(globalconf.L, 1);
-
-            /* Re-push wibox */
-            luaA_object_push(globalconf.L, wibox);
-            luaA_object_unref_item(globalconf.L, -1, wibox->mouse_over);
-            /* Remove the wibox */
-            lua_pop(globalconf.L, 1);
-        }
-        if(widget)
-        {
-            /* Get a ref on this widget so that it can't be unref'd from
-             * underneath us (-> invalid pointer dereference):
-             * - Push the wibox */
-            luaA_object_push(globalconf.L, wibox);
-            /* - Push the widget */
-            luaA_object_push_item(globalconf.L, -1, widget);
-            /* - Reference the widget into the wibox */
-            luaA_object_ref_item(globalconf.L, -2, -1);
-
-            /* Emit mouse::enter signal on new widget:
-             * - Push the widget */
-            luaA_object_push_item(globalconf.L, -1, widget);
-            /* - Move the widget before the wibox we pushed just above */
-            lua_insert(globalconf.L, -2);
-            /* - Emit the signal with the wibox as argument */
-
-            luaA_object_emit_signal(globalconf.L, -2, "mouse::enter", 1);
-            /* - Remove the widget from the stack */
-            lua_pop(globalconf.L, 1);
-        }
-
-        /* Store that this widget was the one with the mouse over */
-        wibox->mouse_over = widget;
-
-    }
-}
-
 /** The motion notify event handler.
  * \param ev The event.
  */
 static void
 event_handle_motionnotify(xcb_motion_notify_event_t *ev)
 {
-    wibox_t *wibox;
+    drawin_t *w;
+    client_t *c;
 
     globalconf.timestamp = ev->time;
 
     if(event_handle_mousegrabber(ev->root_x, ev->root_y, ev->state))
         return;
 
-    if((wibox = wibox_getbywin(ev->event)))
+    if((c = client_getbyframewin(ev->event)))
     {
-        widget_t *w = widget_getbycoords(wibox->orientation, &wibox->widgets,
-                                         wibox->geometry.width,
-                                         wibox->geometry.height,
-                                         &ev->event_x, &ev->event_y);
+        luaA_object_push(globalconf.L, c);
+        lua_pushnumber(globalconf.L, ev->event_x);
+        lua_pushnumber(globalconf.L, ev->event_y);
+        luaA_object_emit_signal(globalconf.L, -3, "mouse::move", 2);
 
-        event_handle_widget_motionnotify(wibox, w);
+        /* now check if a titlebar was "hit" */
+        int x = ev->event_x, y = ev->event_y;
+        drawable_t *d = client_get_drawable_offset(c, &x, &y);
+        if (d)
+        {
+            luaA_object_push_item(globalconf.L, -1, d);
+            lua_pushnumber(globalconf.L, x);
+            lua_pushnumber(globalconf.L, y);
+            luaA_object_emit_signal(globalconf.L, -3, "mouse::move", 2);
+            lua_pop(globalconf.L, 1);
+        }
+        lua_pop(globalconf.L, 1);
+    }
+
+    if((w = drawin_getbywin(ev->event)))
+    {
+        luaA_object_push(globalconf.L, w);
+        luaA_object_push_item(globalconf.L, -1, w->drawable);
+        lua_pushnumber(globalconf.L, ev->event_x);
+        lua_pushnumber(globalconf.L, ev->event_y);
+        luaA_object_emit_signal(globalconf.L, -3, "mouse::move", 2);
+        lua_pop(globalconf.L, 2);
     }
 }
 
@@ -424,7 +396,7 @@ event_handle_motionnotify(xcb_motion_notify_event_t *ev)
 static void
 event_handle_leavenotify(xcb_leave_notify_event_t *ev)
 {
-    wibox_t *wibox;
+    drawin_t *drawin;
     client_t *c;
 
     globalconf.timestamp = ev->time;
@@ -432,39 +404,26 @@ event_handle_leavenotify(xcb_leave_notify_event_t *ev)
     if(ev->mode != XCB_NOTIFY_MODE_NORMAL)
         return;
 
-    if((c = client_getbytitlebarwin(ev->event)) || (c = client_getbywin(ev->event)))
+    if((c = client_getbyframewin(ev->event)))
     {
-        if(globalconf.hooks.mouse_leave != LUA_REFNIL)
-        {
-            luaA_object_push(globalconf.L, c);
-            luaA_dofunction_from_registry(globalconf.L, globalconf.hooks.mouse_leave, 1, 0);
-        }
         luaA_object_push(globalconf.L, c);
         luaA_object_emit_signal(globalconf.L, -1, "mouse::leave", 0);
+        drawable_t *d = client_get_drawable(c, ev->event_x, ev->event_y);
+        if (d)
+        {
+            luaA_object_push_item(globalconf.L, -1, d);
+            luaA_object_emit_signal(globalconf.L, -1, "mouse::leave", 0);
+            lua_pop(globalconf.L, 1);
+        }
         lua_pop(globalconf.L, 1);
     }
 
-    if((wibox = wibox_getbywin(ev->event)))
+    if((drawin = drawin_getbywin(ev->event)))
     {
-        if(wibox->mouse_over)
-        {
-            /* Push the wibox */
-            luaA_object_push(globalconf.L, wibox);
-            /* Push the widget the mouse is over in this wibox */
-            luaA_object_push_item(globalconf.L, -1, wibox->mouse_over);
-            /* Move the widget before the wibox */
-            lua_insert(globalconf.L, -2);
-            /* Emit mouse::leave signal on widget the mouse was over with
-             * its wibox as argument */
-            luaA_object_emit_signal(globalconf.L, -2, "mouse::leave", 1);
-            /* Remove the widget from the stack */
-            lua_pop(globalconf.L, 1);
-            wibox_clear_mouse_over(wibox);
-        }
-
-        luaA_object_push(globalconf.L, wibox);
+        luaA_object_push(globalconf.L, drawin);
+        luaA_object_push_item(globalconf.L, -1, drawin->drawable);
         luaA_object_emit_signal(globalconf.L, -1, "mouse::leave", 0);
-        lua_pop(globalconf.L, 1);
+        lua_pop(globalconf.L, 2);
     }
 }
 
@@ -475,37 +434,32 @@ static void
 event_handle_enternotify(xcb_enter_notify_event_t *ev)
 {
     client_t *c;
-    wibox_t *wibox;
+    drawin_t *drawin;
 
     globalconf.timestamp = ev->time;
 
     if(ev->mode != XCB_NOTIFY_MODE_NORMAL)
         return;
 
-    if((wibox = wibox_getbywin(ev->event)))
+    if((drawin = drawin_getbywin(ev->event)))
     {
-        widget_t *w = widget_getbycoords(wibox->orientation, &wibox->widgets,
-                                         wibox->geometry.width,
-                                         wibox->geometry.height,
-                                         &ev->event_x, &ev->event_y);
-
-        event_handle_widget_motionnotify(wibox, w);
-
-        luaA_object_push(globalconf.L, wibox);
+        luaA_object_push(globalconf.L, drawin);
+        luaA_object_push_item(globalconf.L, -1, drawin->drawable);
         luaA_object_emit_signal(globalconf.L, -1, "mouse::enter", 0);
-        lua_pop(globalconf.L, 1);
+        lua_pop(globalconf.L, 2);
     }
 
-    if((c = client_getbytitlebarwin(ev->event))
-       || (c = client_getbywin(ev->event)))
+    if((c = client_getbyframewin(ev->event)))
     {
-        if(globalconf.hooks.mouse_enter != LUA_REFNIL)
-        {
-            luaA_object_push(globalconf.L, c);
-            luaA_dofunction_from_registry(globalconf.L, globalconf.hooks.mouse_enter, 1, 0);
-        }
         luaA_object_push(globalconf.L, c);
         luaA_object_emit_signal(globalconf.L, -1, "mouse::enter", 0);
+        drawable_t *d = client_get_drawable(c, ev->event_x, ev->event_y);
+        if (d)
+        {
+            luaA_object_push_item(globalconf.L, -1, d);
+            luaA_object_emit_signal(globalconf.L, -1, "mouse::enter", 0);
+            lua_pop(globalconf.L, 1);
+        }
         lua_pop(globalconf.L, 1);
     }
 }
@@ -516,7 +470,10 @@ event_handle_enternotify(xcb_enter_notify_event_t *ev)
 static void
 event_handle_focusin(xcb_focus_in_event_t *ev)
 {
-    client_t *c;
+    if (ev->mode == XCB_NOTIFY_MODE_GRAB
+            || ev->mode == XCB_NOTIFY_MODE_UNGRAB)
+        /* Ignore focus changes due to keyboard grabs */
+        return;
 
     /* Events that we are interested in: */
     switch(ev->detail)
@@ -531,9 +488,15 @@ event_handle_focusin(xcb_focus_in_event_t *ev)
          */
         case XCB_NOTIFY_DETAIL_NONLINEAR_VIRTUAL:
         case XCB_NOTIFY_DETAIL_NONLINEAR:
-            if((c = client_getbytitlebarwin(ev->event))
-               || (c = client_getbywin(ev->event)))
+          {
+            client_t *c;
+
+            if((c = client_getbywin(ev->event))) {
+                /* If there is still a pending focus change, do it now. */
+                client_focus_refresh();
                 client_focus_update(c);
+            }
+          }
         /* all other events are ignored */
         default:
             break;
@@ -546,15 +509,15 @@ event_handle_focusin(xcb_focus_in_event_t *ev)
 static void
 event_handle_expose(xcb_expose_event_t *ev)
 {
-    wibox_t *wibox;
+    drawin_t *drawin;
+    client_t *client;
 
-    /* If the wibox got need_update set, skip this because it will be repainted
-     * soon anyway. Without this we could be painting garbage to the screen!
-     */
-    if((wibox = wibox_getbywin(ev->window)) && !wibox->need_update)
-        wibox_refresh_pixmap_partial(wibox,
-                                     ev->x, ev->y,
-                                     ev->width, ev->height);
+    if((drawin = drawin_getbywin(ev->window)))
+        drawin_refresh_pixmap_partial(drawin,
+                                      ev->x, ev->y,
+                                      ev->width, ev->height);
+    if ((client = client_getbyframewin(ev->window)))
+        client_refresh(client);
 }
 
 /** The key press event handler.
@@ -575,17 +538,15 @@ event_handle_key(xcb_key_press_event_t *ev)
                 warn("error running function: %s", lua_tostring(globalconf.L, -1));
                 luaA_keygrabber_stop(globalconf.L);
             }
-            else if(!lua_isboolean(globalconf.L, -1) || !lua_toboolean(globalconf.L, -1))
-                luaA_keygrabber_stop(globalconf.L);
         }
         lua_pop(globalconf.L, 1);  /* pop returned value or function if not called */
     }
     else
     {
         /* get keysym ignoring all modifiers */
-        xcb_keysym_t keysym = key_getkeysym(ev->detail, 0);
+        xcb_keysym_t keysym = keyresolv_get_keysym(ev->detail, 0);
         client_t *c;
-        if((c = client_getbywin(ev->event)))
+        if((c = client_getbyframewin(ev->event)))
         {
             luaA_object_push(globalconf.L, c);
             event_key_callback(ev, &c->keys, -1, 1, &keysym);
@@ -601,7 +562,6 @@ event_handle_key(xcb_key_press_event_t *ev)
 static void
 event_handle_maprequest(xcb_map_request_event_t *ev)
 {
-    int phys_screen;
     client_t *c;
     xcb_get_window_attributes_cookie_t wa_c;
     xcb_get_window_attributes_reply_t *wa_r;
@@ -624,7 +584,7 @@ event_handle_maprequest(xcb_map_request_event_t *ev)
     else if((c = client_getbywin(ev->window)))
     {
         /* Check that it may be visible, but not asked to be hidden */
-        if(client_maybevisible(c, c->screen) && !c->hidden)
+        if(client_maybevisible(c) && !c->hidden)
         {
             luaA_object_push(globalconf.L, c);
             client_set_minimized(globalconf.L, -1, false);
@@ -642,9 +602,7 @@ event_handle_maprequest(xcb_map_request_event_t *ev)
             goto bailout;
         }
 
-        phys_screen = xutil_root2screen(globalconf.connection, geom_r->root);
-
-        client_manage(ev->window, geom_r, phys_screen, false);
+        client_manage(ev->window, geom_r, false);
 
         p_delete(&geom_r);
     }
@@ -662,21 +620,14 @@ event_handle_unmapnotify(xcb_unmap_notify_event_t *ev)
     client_t *c;
 
     if((c = client_getbywin(ev->window)))
-    {
-        if(ev->event == xutil_screen_get(globalconf.connection, c->phys_screen)->root
-           && XCB_EVENT_SENT(ev))
-        {
-            client_unmanage(c);
-            xcb_unmap_window(globalconf.connection, ev->window);
-        }
-    }
+        client_unmanage(c, true);
     else
         for(int i = 0; i < globalconf.embedded.len; i++)
             if(globalconf.embedded.tab[i].win == ev->window)
             {
                 xembed_window_array_take(&globalconf.embedded, i);
-                widget_invalidate_bytype(widget_systray);
                 xcb_change_save_set(globalconf.connection, XCB_SET_MODE_DELETE, ev->window);
+                luaA_systray_invalidate();
             }
 }
 
@@ -757,22 +708,17 @@ event_handle_mappingnotify(xcb_mapping_notify_event_t *ev)
                             &globalconf.shiftlockmask, &globalconf.capslockmask,
                             &globalconf.modeswitchmask);
 
-        int nscreen = xcb_setup_roots_length(xcb_get_setup(globalconf.connection));
-
         /* regrab everything */
-        for(int phys_screen = 0; phys_screen < nscreen; phys_screen++)
-        {
-            xcb_screen_t *s = xutil_screen_get(globalconf.connection, phys_screen);
-            /* yes XCB_BUTTON_MASK_ANY is also for grab_key even if it's look weird */
-            xcb_ungrab_key(globalconf.connection, XCB_GRAB_ANY, s->root, XCB_BUTTON_MASK_ANY);
-            window_grabkeys(s->root, &globalconf.keys);
-        }
+        xcb_screen_t *s = globalconf.screen;
+        /* yes XCB_BUTTON_MASK_ANY is also for grab_key even if it's look weird */
+        xcb_ungrab_key(globalconf.connection, XCB_GRAB_ANY, s->root, XCB_BUTTON_MASK_ANY);
+        xwindow_grabkeys(s->root, &globalconf.keys);
 
         foreach(_c, globalconf.clients)
         {
             client_t *c = *_c;
-            xcb_ungrab_key(globalconf.connection, XCB_GRAB_ANY, c->window, XCB_BUTTON_MASK_ANY);
-            window_grabkeys(c->window, &c->keys);
+            xcb_ungrab_key(globalconf.connection, XCB_GRAB_ANY, c->frame_window, XCB_BUTTON_MASK_ANY);
+            xwindow_grabkeys(c->frame_window, &c->keys);
         }
     }
 }
@@ -782,8 +728,13 @@ event_handle_reparentnotify(xcb_reparent_notify_event_t *ev)
 {
     client_t *c;
 
-    if((c = client_getbywin(ev->window)))
-        client_unmanage(c);
+    if((c = client_getbywin(ev->window)) && c->frame_window != ev->parent)
+    {
+        /* Ignore reparents to the root window, they *might* be caused by
+         * ourselves if a client quickly unmaps and maps itself again. */
+        if (ev->parent != globalconf.screen->root)
+            client_unmanage(c, true);
+    }
 }
 
 /** \brief awesome xerror function.
@@ -800,13 +751,15 @@ xerror(xcb_generic_error_t *e)
            && e->major_code == XCB_SET_INPUT_FOCUS)
        || (e->error_code == XCB_VALUE
            && e->major_code == XCB_KILL_CLIENT)
-       || (e->major_code == XCB_CONFIGURE_WINDOW
-           && e->error_code == XCB_MATCH))
+       || (e->error_code == XCB_MATCH
+           && e->major_code == XCB_CONFIGURE_WINDOW))
         return;
 
-    warn("X error: request=%s, error=%s",
+    warn("X error: request=%s (major %d, minor %d), error=%s (%d)",
          xcb_event_get_request_label(e->major_code),
-         xcb_event_get_error_label(e->error_code));
+         e->major_code, e->minor_code,
+         xcb_event_get_error_label(e->error_code),
+         e->error_code);
 
     return;
 }
@@ -861,4 +814,4 @@ void event_handle(xcb_generic_event_t *event)
         event_handle_randr_screen_change_notify((void *) event);
 }
 
-// vim: filetype=c:expandtab:shiftwidth=4:tabstop=8:softtabstop=4:encoding=utf-8:textwidth=80
+// vim: filetype=c:expandtab:shiftwidth=4:tabstop=8:softtabstop=4:textwidth=80
